#=========================================================================
# TranslationPass.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 12, 2019
"""Translate a PyMTL component hierarhcy into SystemVerilog source code."""

import filecmp
import os

from pymtl3.passes.backends.verilog import TranslationConfigs
from pymtl3.passes.BasePass import BasePass, PassMetadata

from .VTranslator import VTranslator


def mk_TranslationPass( _VTranslator ):

  class _TranslationPass( BasePass ):

    def __call__( s, top ):

      s.top = top
      s.translator = _VTranslator( s.top )
      s.traverse_hierarchy( top )

    def gen_tr_cfgs( s, m ):
      tr_cfgs = {}

      def traverse( m ):
        nonlocal tr_cfgs

        if not hasattr( m, 'config_verilog_translate' ) or \
           isinstance( m.config_verilog_translate, bool ):
          tr_cfgs[m] = TranslationConfigs()
        else:
          tr_cfgs[m] = m.config_verilog_translate

        for _m in m.get_child_components(repr):
          traverse( _m )

      traverse( m )
      return tr_cfgs

    def traverse_hierarchy( s, m ):

      if hasattr(m, "config_verilog_translate") and m.config_verilog_translate and \
         m.config_verilog_translate.translate:

        if not hasattr( m, '_pass_verilog_translation' ):
          m._pass_verilog_translation = PassMetadata()

        s.translator.translate( m, s.gen_tr_cfgs(m) )

        module_name = s.translator._top_module_full_name

        if m.config_verilog_translate.explicit_file_name:
          fname = m.config_verilog_translate.explicit_file_name
          if '.v' in fname:
            filename = fname.split('.v')[0]
          elif '.sv' in fname:
            filename = fname.split('.sv')[0]
          else:
            filename = fname
        else:
          filename = module_name

        output_file = filename + '.v'
        temporary_file = filename + '.v.tmp'

        # First write the file to a temporary file
        m._pass_verilog_translation.is_same = False
        with open( temporary_file, 'w' ) as output:
          output.write( s.translator.hierarchy.src )
          output.flush()
          os.fsync( output )
          output.close()

        # `is_same` is set if there exists a file that has the same filename as
        # `output_file`, and that file is the same as the temporary file
        if ( os.path.exists(output_file) ):
          m._pass_verilog_translation.is_same = \
              filecmp.cmp( temporary_file, output_file )

        # Rename the temporary file to the output file
        os.rename( temporary_file, output_file )

        # Expose some attributes about the translation process
        m.translated_top_module_name = module_name
        m._translator = s.translator
        m._pass_verilog_translation.translated = True

        m._pass_verilog_translation.translated_filename = output_file
        m._pass_verilog_translation.translated_top_module = module_name

      else:
        for child in m.get_child_components(repr):
          s.traverse_hierarchy( child )

  return _TranslationPass

TranslationPass = mk_TranslationPass( VTranslator )



#=========================================================================
# TranslationImportPass.py
#=========================================================================
# Translate and import components having the `verilog_translate_import`
# attribute.
#
# Author : Peitian Pan
# Date   : Aug 6, 2019

from pymtl3 import Placeholder
from pymtl3.passes.BasePass import BasePass

from .import_.VerilatorImportConfigs import VerilatorImportConfigs
from .import_.VerilatorImportPass import VerilatorImportPass
from .translation.TranslationConfigs import TranslationConfigs
from .translation.TranslationPass import TranslationPass
from .VerilogPlaceholderConfigs import VerilogPlaceholderConfigs


class TranslationImportPass( BasePass ):

  def __call__( s, top ):
    s.top = top
    s.traverse_hierarchy( top )
    top.apply( s.get_translation_pass() )
    s.add_placeholder_marks( top )
    return s.get_import_pass()( top )

  def traverse_hierarchy( s, m ):
    search_subtree = True

    # Found a subtree that is marked as to be translated and imported
    if hasattr(m, s.get_flag_name()) and getattr(m, s.get_flag_name()):

      # Use default translation config if it is missing
      if not hasattr(m, s.get_translation_flag_name()):
        setattr(m, s.get_translation_flag_name(), s.get_translation_configs())

      # Assert that this instance should be translated
      getattr(m, s.get_translation_flag_name()).translate = True

      if not hasattr(m, s.get_import_flag_name()):
        setattr(m, s.get_import_flag_name(), s.get_import_configs())

      search_subtree = False

    if search_subtree:
      for child in m.get_child_components(repr):
        s.traverse_hierarchy( child )

  def add_placeholder_marks( s, m ):
    # TODO: what we really want is to generate placeholders for components
    # to be translated and imported. Right now we simply attach the placeholder
    # and import configs to the components.
    if hasattr( m, s.get_flag_name() ) and getattr( m, s.get_flag_name() ):

      if not hasattr( m, 'config_placeholder' ):
        # If m is not a placeholder, we need to populate the v_include
        # placeholder config from all submodules into m.
        vi = list(s.get_hierarchy_v_include(m))
        m.config_placeholder = VerilogPlaceholderConfigs(v_include=vi)

      if not hasattr( m, s.get_import_flag_name() ):
        m.config_verilog_import = VerilatorImportConfigs()

      m.config_placeholder.pickled_dependency_file = None
      m.config_placeholder.pickled_wrapper_file = \
          getattr(m, s.get_translation_pass_namespace()).translated_filename
      m.config_placeholder.pickled_top_module = \
          getattr(m, s.get_translation_pass_namespace()).translated_top_module

    else:
      for child in m.get_child_components(repr):
        s.add_placeholder_marks( child )

  def get_translation_pass( s ):
    return TranslationPass()

  def get_import_pass( s ):
    return VerilatorImportPass()

  def get_flag_name( s ):
    return "verilog_translate_import"

  def get_translation_flag_name( s ):
    return "config_verilog_translate"

  def get_translation_pass_namespace( s ):
    return "_pass_verilog_translation"

  def get_import_flag_name( s ):
    return "config_verilog_import"

  def get_translation_configs( s ):
    return TranslationConfigs()

  def get_import_configs( s ):
    return VerilatorImportConfigs(vl_Wno_list=['UNOPTFLAT', 'UNSIGNED', 'WIDTH'])

  def get_hierarchy_v_include( s, m ):
    all_v_includes = set()
    if isinstance( m, Placeholder ) and hasattr( m, 'config_placeholder' ):
      for v_include in m.config_placeholder.v_include:
        all_v_includes.add( v_include )
    else:
      for child in m.get_child_components(repr):
        all_v_includes |= s.get_hierarchy_v_include( child )
    return all_v_includes



#=========================================================================
# ImportedObject_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 2, 2019
"""Test if the imported object works correctly."""

from os.path import dirname

import pytest

from pymtl3 import SimulationPass
from pymtl3.datatypes import Bits1, Bits32, Bits48, Bits64, clog2, mk_bits
from pymtl3.dsl import Component, InPort, Interface, OutPort, Placeholder, connect
from pymtl3.passes.backends.verilog import (
    TranslationConfigs,
    TranslationImportPass,
    VerilatorImportConfigs,
    VerilogPlaceholderConfigs,
    VerilogPlaceholderPass,
)
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.stdlib.test import TestVectorSimulator


def local_do_test( _m ):
  _m.elaborate()
  _m.apply( VerilogPlaceholderPass() )
  m = TranslationImportPass()( _m )
  sim = TestVectorSimulator( m, _m._test_vectors, _m._tv_in, _m._tv_out )
  sim.run_test()

def test_reg( do_test ):
  # General trans-import test
  def tv_in( m, test_vector ):
    m.in_ = Bits32( test_vector[0] )
  def tv_out( m, test_vector ):
    if test_vector[1] != '*':
      assert m.out == Bits32( test_vector[1] )
  class VReg( Component, Placeholder ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.config_placeholder = VerilogPlaceholderConfigs(
          src_file = dirname(__file__)+'/VReg.v',
          port_map = {
            "clk" : "clk",
            "reset" : "reset",
            "in_" : "d",
            "out" : "q",
          }
      )
      s.config_verilog_translate = TranslationConfigs(
          explicit_module_name = 'VReg_imported',
      )
      s.verilog_translate_import = True
  a = VReg()
  a._test_vectors = [
    [    1,    '*' ],
    [    2,      1 ],
    [   -1,      2 ],
    [   -2,     -1 ],
    [   42,     -2 ],
    [  -42,     42 ],
  ]
  a._tv_in = tv_in
  a._tv_out = tv_out
  do_test( a )

def test_vl_uninit( do_test ):
  # Use a latch to test if verilator has correctly set up
  # the inital signal values
  def tv_in( m, test_vector ):
    m.in_ = Bits32( test_vector[0] )
  def tv_out( m, test_vector ):
    assert m.out == Bits32( test_vector[1] )
  class Uninit( Component, Placeholder ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.config_placeholder = VerilogPlaceholderConfigs(
          src_file = dirname(__file__)+'/VUninit.v',
          top_module = 'VUninit',
          port_map = {
            "in_" : "d",
            "out" : "q",
          }
      )
      s.config_verilog_import = VerilatorImportConfigs(
          vl_xinit = 'ones',
      )
      s.verilog_translate_import = True
  a = Uninit()
  a._test_vectors = [
    [    0, 4294967295 ],
    [    2, 4294967295 ],
    [   42,         42 ],
  ]
  a._tv_in = tv_in
  a._tv_out = tv_out
  do_test( a )

def test_reg_external_trace( do_test ):
  # Test Verilog line trace
  class RegTrace( Component, Placeholder ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.config_placeholder = VerilogPlaceholderConfigs(
          src_file = dirname(__file__)+'/VRegTrace.v',
          top_module = 'VRegTrace',
          port_map = {
            "in_" : "d",
            "out" : "q",
          }
      )
      s.config_verilog_import = VerilatorImportConfigs(
          vl_line_trace = True,
      )
      s.verilog_translate_import = True
  a = RegTrace()
  a.elaborate()
  a.apply( VerilogPlaceholderPass() )
  a = TranslationImportPass()( a )
  a.apply( SimulationPass() )

  a.in_ = Bits32(1)
  a.tick()
  assert a.line_trace() == 'q =          0'
  a.in_ = Bits32(2)
  a.tick()
  assert a.line_trace() == 'q =          1'
  a.in_ = Bits32(-1)
  a.tick()
  assert a.line_trace() == 'q =          2'
  a.tick()
  # 0xFFFFFFFF unsigned
  assert a.line_trace() == 'q = 4294967295'

def test_reg_incomplete_portmap( do_test ):
  # Test support for incomplete port map
  def tv_in( m, test_vector ):
    m.in_ = Bits32( test_vector[0] )
  def tv_out( m, test_vector ):
    if test_vector[1] != '*':
      assert m.out == Bits32( test_vector[1] )
  class Reg( Component, Placeholder ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.config_placeholder = VerilogPlaceholderConfigs(
          src_file = dirname(__file__)+'/VReg.v',
          top_module = 'VReg',
          port_map = {
            "in_" : "d",
            "out" : "q",
          }
      )
      s.verilog_translate_import = True
  a = Reg()
  a._test_vectors = [
    [    1,    '*' ],
    [    2,      1 ],
    [   -1,      2 ],
    [   -2,     -1 ],
    [   42,     -2 ],
    [  -42,     42 ],
  ]
  a._tv_in = tv_in
  a._tv_out = tv_out
  do_test( a )

def test_adder( do_test ):
  # Test an adder
  def tv_in( m, test_vector ):
    m.in0 = Bits32( test_vector[0] )
    m.in1 = Bits32( test_vector[1] )
    m.cin = Bits1( test_vector[2] )
  def tv_out( m, test_vector ):
    if test_vector[3] != '*':
      assert m.out == Bits32( test_vector[3] )
    if test_vector[4] != '*':
      assert m.cout == Bits32( test_vector[4] )
  class Adder( Component, Placeholder ):
    def construct( s ):
      s.in0 = InPort( Bits32 )
      s.in1 = InPort( Bits32 )
      s.cin = InPort( Bits1 )
      s.out = OutPort( Bits32 )
      s.cout = OutPort( Bits1 )
      s.config_placeholder = VerilogPlaceholderConfigs(
          src_file = dirname(__file__)+'/VAdder.v',
          top_module = 'VAdder',
      )
      s.verilog_translate_import = True
  a = Adder()
  a._test_vectors = [
    [    1,      1,     1,     3, 0 ],
    [    1,     -1,     0,     0, 1 ],
    [   42,     42,     1,    85, 0 ],
    [   42,    -43,     1,     0, 1 ],
  ]
  a._tv_in = tv_in
  a._tv_out = tv_out
  do_test( a )

def test_normal_queue( do_test ):
  # Test a Placeholder with params in `construct`
  def tv_in( m, tv ):
    m.enq_en = Bits1( tv[0] )
    m.enq_msg = Bits32( tv[1] )
    m.deq_en = Bits1( tv[3] )
  def tv_out( m, tv ):
    if tv[2] != '*':
      assert m.enq_rdy == Bits1( tv[2] )
    if tv[4] != '*':
      assert m.deq_rdy == Bits1( tv[5] )
    if tv[5] != '*':
      assert m.deq_msg == Bits32( tv[4] )
  class Queue( Component, Placeholder ):
    def construct( s, data_width, num_entries, count_width ):
      s.count   =  OutPort( mk_bits( count_width )  )
      s.deq_en  =  InPort( Bits1  )
      s.deq_rdy = OutPort( Bits1  )
      s.deq_msg = OutPort( mk_bits( data_width ) )
      s.enq_en  =  InPort( Bits1  )
      s.enq_rdy = OutPort( Bits1  )
      s.enq_msg =  InPort( mk_bits( data_width ) )
      s.config_placeholder = VerilogPlaceholderConfigs(
          src_file = dirname(__file__)+'/VQueue.v',
          top_module = 'VQueue',
      )
      s.verilog_translate_import = True
  num_entries = 1
  q = Queue(
      data_width = 32,
      num_entries = num_entries,
      count_width = clog2(num_entries+1))
  # q.dump_vcd = True
  test_vector = [
    #   enq                deq
    #   en    msg   rdy    en    msg   rdy
    [    1,    42,    1,    0,     0,    0  ],
    [    0,    43,    0,    1,    42,    1  ],
    [    1,    43,    1,    0,    42,    0  ],
    [    0,    44,    0,    1,    43,    1  ],
    [    1,    44,    1,    0,    43,    0  ],
    [    0,    45,    0,    1,    44,    1  ],
    [    1,    45,    1,    0,    44,    0  ],
  ]
  q._test_vectors = test_vector
  q._tv_in = tv_in
  q._tv_out = tv_out
  do_test( q )

def test_normal_queue_implicit_top_module( do_test ):
  # Test a Placeholder with params in `construct`
  def tv_in( m, tv ):
    m.enq_en = Bits1( tv[0] )
    m.enq_msg = Bits32( tv[1] )
    m.deq_en = Bits1( tv[3] )
  def tv_out( m, tv ):
    if tv[2] != '*':
      assert m.enq_rdy == Bits1( tv[2] )
    if tv[4] != '*':
      assert m.deq_rdy == Bits1( tv[5] )
    if tv[5] != '*':
      assert m.deq_msg == Bits32( tv[4] )
  class VQueue( Component, Placeholder ):
    def construct( s, data_width, num_entries, count_width ):
      s.count   =  OutPort( mk_bits( count_width )  )
      s.deq_en  =  InPort( Bits1  )
      s.deq_rdy = OutPort( Bits1  )
      s.deq_msg = OutPort( mk_bits( data_width ) )
      s.enq_en  =  InPort( Bits1  )
      s.enq_rdy = OutPort( Bits1  )
      s.enq_msg =  InPort( mk_bits( data_width ) )
      s.config_placeholder = VerilogPlaceholderConfigs(
          src_file = dirname(__file__)+'/VQueue.v',
      )
      s.verilog_translate_import = True
  num_entries = 1
  q = VQueue(
      data_width = 32,
      num_entries = num_entries,
      count_width = clog2(num_entries+1))
  # q.dump_vcd = True
  test_vector = [
    #   enq                deq
    #   en    msg   rdy    en    msg   rdy
    [    1,    42,    1,    0,     0,    0  ],
    [    0,    43,    0,    1,    42,    1  ],
    [    1,    43,    1,    0,    42,    0  ],
    [    0,    44,    0,    1,    43,    1  ],
    [    1,    44,    1,    0,    43,    0  ],
    [    0,    45,    0,    1,    44,    1  ],
    [    1,    45,    1,    0,    44,    0  ],
  ]
  q._test_vectors = test_vector
  q._tv_in = tv_in
  q._tv_out = tv_out
  do_test( q )

def test_normal_queue_params( do_test ):
  # Test the `params` option of placeholder configs
  def tv_in( m, tv ):
    m.enq_en = Bits1( tv[0] )
    m.enq_msg = Bits32( tv[1] )
    m.deq_en = Bits1( tv[3] )
  def tv_out( m, tv ):
    if tv[2] != '*':
      assert m.enq_rdy == Bits1( tv[2] )
    if tv[4] != '*':
      assert m.deq_rdy == Bits1( tv[5] )
    if tv[5] != '*':
      assert m.deq_msg == Bits32( tv[4] )
  class Queue( Component, Placeholder ):
    def construct( s, nbits, nelems, nbits_cnt ):
      s.count   =  OutPort( mk_bits( nbits_cnt )  )
      s.deq_en  =  InPort( Bits1  )
      s.deq_rdy = OutPort( Bits1  )
      s.deq_msg = OutPort( mk_bits( nbits ) )
      s.enq_en  =  InPort( Bits1  )
      s.enq_rdy = OutPort( Bits1  )
      s.enq_msg =  InPort( mk_bits( nbits ) )
      s.config_placeholder = VerilogPlaceholderConfigs(
          src_file = dirname(__file__)+'/VQueue.v',
          top_module = 'VQueue',
          params = {
            'data_width'  : nbits,
            'num_entries' : nelems,
            'count_width' : nbits_cnt,
          },
      )
      s.verilog_translate_import = True
  num_entries = 1
  q = Queue(
      nbits = 32,
      nelems = num_entries,
      nbits_cnt = clog2(num_entries+1))
  # q.dump_vcd = True
  test_vector = [
    #   enq                deq
    #   en    msg   rdy    en    msg   rdy
    [    1,    42,    1,    0,     0,    0  ],
    [    0,    43,    0,    1,    42,    1  ],
    [    1,    43,    1,    0,    42,    0  ],
    [    0,    44,    0,    1,    43,    1  ],
    [    1,    44,    1,    0,    43,    0  ],
    [    0,    45,    0,    1,    44,    1  ],
    [    1,    45,    1,    0,    44,    0  ],
  ]
  q._test_vectors = test_vector
  q._tv_in = tv_in
  q._tv_out = tv_out
  do_test( q )

def test_unpacked_port_array( do_test ):
  # Test the `params` option of placeholder configs
  def tv_in( m, tv ):
    m.in_[0] = Bits32(tv[0])
    m.in_[1] = Bits32(tv[1])
  def tv_out( m, tv ):
    assert m.out[0] == Bits32(tv[2])
    assert m.out[1] == Bits32(tv[3])
  class VPassThrough( Component, Placeholder ):
    def construct( s, nports, nbits ):
      s.in_ = [ InPort( mk_bits(nbits) ) for _ in range(nports) ]
      s.out = [ OutPort( mk_bits(nbits) ) for _ in range(nports) ]

      s.config_placeholder = VerilogPlaceholderConfigs(
          src_file = dirname(__file__) + '/VPassThrough.v',
          params = {
            'num_ports' : nports,
            'bitwidth'  : nbits,
          },
          has_clk = False,
          has_reset = False,
      )
      s.verilog_translate_import = True

  q = VPassThrough( 2, 32 )
  test_vector = [
    [ 1, 42, 1, 42 ],
    [ 1, -1, 1, -1 ],
    [ 0,  1, 0,  1 ],
    [ -1, 1, -1, 1 ],
  ]
  q._test_vectors = test_vector
  q._tv_in = tv_in
  q._tv_out = tv_out
  do_test( q )

@pytest.mark.parametrize(
  "translate", [ True, False ]
)
def test_param_pass_through( do_test, translate ):
  class VPassThrough( Component, Placeholder ):
    def construct( s, nports, nbits ):
      s.in_ = [ InPort( mk_bits(nbits) ) for _ in range(nports) ]
      s.out = [ OutPort( mk_bits(nbits) ) for _ in range(nports) ]
      s.config_placeholder = VerilogPlaceholderConfigs(
          src_file = dirname(__file__) + '/VPassThrough.v',
          params = {
            'num_ports' : nports,
            'bitwidth'  : nbits,
          },
          has_clk = False,
          has_reset = False,
      )
      s.verilog_translate_import = True
  class PassThrough( Component ):
    def construct( s ):
      s.in_ = InPort( Bits48 )
      s.out = OutPort( Bits48 )
      s.pt16 = VPassThrough(1, 16)
      s.pt32 = VPassThrough(1, 32)
      s.pt16.in_[0] //= s.in_[0:16]
      s.pt16.out[0] //= s.out[0:16]
      s.pt32.in_[0] //= s.in_[16:48]
      s.pt32.out[0] //= s.out[16:48]
      if translate:
        s.verilog_translate_import = True
    def line_trace( s ):
      return f"{s.in_} > {s.out}"
  def tv_in( m, tv ):
    m.in_ = Bits48( tv[0] )
  def tv_out( m, tv ):
    assert m.out == Bits48( tv[1] )

  p = PassThrough()
  test_vector = [
    [  1,  1, ],
    [ -1, -1, ],
    [ 42, 42, ],
    [ -2, -2, ],
  ]
  p._test_vectors = test_vector
  p._tv_in = tv_in
  p._tv_out = tv_out

def test_non_top_portmap( do_test ):
  def tv_in( m, tv ):
    m.in_ = Bits32(tv[0])
  def tv_out( m, tv ):
    if tv[1] != '*':
      assert m.out == Bits32(tv[1])
  class VReg( Component, Placeholder ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.config_placeholder = VerilogPlaceholderConfigs(
          src_file = dirname(__file__)+'/VReg.v',
          port_map = {
            "clk" : "clk",
            "reset" : "reset",
            "in_" : "d",
            "out" : "q",
          }
      )
      s.config_verilog_translate = TranslationConfigs(
          explicit_module_name = 'VReg_imported',
      )
      s.verilog_translate_import = True
  class Top( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.v = VReg()
      s.v.in_ //= s.in_
      s.v.out //= s.out
      s.verilog_translate_import = True
  a = Top()
  a._test_vectors = [
    [    1,    '*' ],
    [    2,      1 ],
    [   -1,      2 ],
    [   -2,     -1 ],
    [   42,     -2 ],
    [  -42,     42 ],
  ]
  a._tv_in = tv_in
  a._tv_out = tv_out
  do_test( a )



#=========================================================================
# ImportedObject_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 2, 2019
"""Test if the imported object works correctly."""

from os.path import dirname

import pytest

from pymtl3 import SimulationPass
from pymtl3.datatypes import Bits1, Bits32, Bits48, Bits64, clog2, mk_bits
from pymtl3.dsl import Component, InPort, Interface, OutPort, Placeholder, connect
from pymtl3.passes.backends.verilog import (
    TranslationConfigs,
    TranslationImportPass,
    VerilatorImportConfigs,
    VerilogPlaceholderConfigs,
    VerilogPlaceholderPass,
    VerilogTBGenPass,
)
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.stdlib.test import TestVectorSimulator

from ...testcases import CaseConnectArrayBits32FooIfcComp


def local_do_test( _m ):
  _m.elaborate()
  _m.apply( VerilogPlaceholderPass() )
  m = TranslationImportPass()( _m )
  m.verilog_tbgen = '1234'
  m.apply( VerilogTBGenPass() )
  sim = TestVectorSimulator( m, _m._test_vectors, _m._tv_in, _m._tv_out )
  sim.run_test()

def test_normal_queue( do_test ):
  # Test a Placeholder with params in `construct`
  def tv_in( m, tv ):
    m.enq_en = Bits1( tv[0] )
    m.enq_msg = Bits32( tv[1] )
    m.deq_en = Bits1( tv[3] )
  def tv_out( m, tv ):
    if tv[2] != '*':
      assert m.enq_rdy == Bits1( tv[2] )
    if tv[4] != '*':
      assert m.deq_rdy == Bits1( tv[5] )
    if tv[5] != '*':
      assert m.deq_msg == Bits32( tv[4] )
  class Queue( Component, Placeholder ):
    def construct( s, data_width, num_entries, count_width ):
      s.count   =  OutPort( mk_bits( count_width )  )
      s.deq_en  =  InPort( Bits1  )
      s.deq_rdy = OutPort( Bits1  )
      s.deq_msg = OutPort( mk_bits( data_width ) )
      s.enq_en  =  InPort( Bits1  )
      s.enq_rdy = OutPort( Bits1  )
      s.enq_msg =  InPort( mk_bits( data_width ) )
      s.config_placeholder = VerilogPlaceholderConfigs(
          src_file = dirname(__file__)+'/VQueue.v',
          top_module = 'VQueue',
      )
      s.verilog_translate_import = True
  num_entries = 1
  q = Queue(
      data_width = 32,
      num_entries = num_entries,
      count_width = clog2(num_entries+1))
  # q.dump_vcd = True
  test_vector = [
    #   enq                deq
    #   en    msg   rdy    en    msg   rdy
    [    1,    42,    1,    0,     0,    0  ],
    [    0,    43,    0,    1,    42,    1  ],
    [    1,    43,    1,    0,    42,    0  ],
    [    0,    44,    0,    1,    43,    1  ],
    [    1,    44,    1,    0,    43,    0  ],
    [    0,    45,    0,    1,    44,    1  ],
    [    1,    45,    1,    0,    44,    0  ],
  ]
  q._test_vectors = test_vector
  q._tv_in = tv_in
  q._tv_out = tv_out
  do_test( q )

def test_CaseConnectArrayBits32FooIfcComp():
  case = CaseConnectArrayBits32FooIfcComp
  try:
    _m = case.DUT()
    _m.elaborate()
    _m.verilog_translate_import = True
    _m.apply( VerilogPlaceholderPass() )
    m = TranslationImportPass()( _m )
    m.verilog_tbgen = True
    m.apply( VerilogTBGenPass() )
    sim = TestVectorSimulator( m, case.TEST_VECTOR, case.TV_IN, case.TV_OUT )
    sim.run_test()
  finally:
    try:
      m.finalize()
    except UnboundLocalError:
      # This test fails due to translation errors
      pass



#=========================================================================
# VerilatorImportConfigs.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jul 28, 2019
"""Configuration of Verilator import pass."""

import os
import subprocess
from textwrap import fill, indent

from pymtl3.passes.errors import InvalidPassOptionValue
from pymtl3.passes.PassConfigs import BasePassConfigs, Checker
from pymtl3.passes.PlaceholderConfigs import expand


class VerilatorImportConfigs( BasePassConfigs ):

  Options = {
    # Enable verbose mode?
    "verbose" : False,

    # Enable external line trace?
    # Once enabled, the `line_trace()` method of the imported component
    # will return a string read from the external `line_trace()` function.
    # This means your Verilog module has to have a `line_trace` function
    # that provides the line trace string which has less than 512 characters.
    # Default to False
    "vl_line_trace" : False,

    # Enable all verilator coverage
    "vl_coverage" : False,

    # Enable all verilator coverage
    "vl_line_coverage" : False,

    # Enable all verilator coverage
    "vl_toggle_coverage" : False,

    # Verilator code generation options
    # These options will be passed to verilator to generate the C simulator.
    # By default, verilator is called with `--cc`.

    # --Mdir
    # Expects the path of Makefile output directory;
    # "" to use `obj_dir_<translated_top_module>`
    "vl_mk_dir" : "",

    # --assert
    # Expects a boolean value
    "vl_enable_assert" : True,

    # Verilator optimization options

    # -O0/3
    # Expects a non-negative integer
    # Currently only support 0 (disable opt) and 3 (highest effort opt)
    "vl_opt_level" : 3,

    # --unroll-count
    # Expects a non-negative integer
    # 0 to disable this option
    "vl_unroll_count" : 1000000,

    # --unroll-stmts
    # Expects a non-negative integer
    # 0 to disable this option
    "vl_unroll_stmts" : 1000000,

    # Verilator warning-related options

    # False to disable the warnings, True to enable
    "vl_W_lint" : True,
    "vl_W_style" : True,
    "vl_W_fatal" : True,

    # Un-warn all warnings in the given list; [] to disable this option
    # The given list should only include strings that appear in `Warnings`
    "vl_Wno_list" : [ 'UNOPTFLAT', 'UNSIGNED', 'WIDTH' ],

    # Verilator misc options

    # What is the inital value of signals?
    # Should be one of ['zeros', 'ones', 'rand']
    "vl_xinit" : "zeros",

    # --trace
    # Expects a boolean value
    "vl_trace" : False,

    # The output filename of Verilator VCD tracing
    # default is {component_name}.verilator1
    "vl_trace_filename" : "",

    # Passed to verilator tracing function
    "vl_trace_timescale" : "10ps",

    # `vl_trace_cycle_time`*`vl_trace_timescale` is the cycle time of the
    # PyMTL clock that appears in the generated VCD
    # With the default options, the frequency of PyMTL clock is 1GHz
    "vl_trace_cycle_time" : 100,

    # C-compilation options
    # These options will be passed to the C compiler to create a shared lib.

    # Additional flags to be passed to the C compiler.
    # By default, CC is called with `-O0 -fPIC -shared`.
    # "" to disable this option
    "c_flags" : "",

    # Additional include search path of the C compiler.
    # [] to disable this option
    "c_include_path" : [],

    # Additional C source files passed to the C compiler.
    # [] to compile verilator generated files only.
    "c_srcs" : [],

    # `LDLIBS` will be listed after the primary target file whereas
    # `LDFLAGS` will be listed before.

    # We enforce the GNU makefile implicit rule that `LDFLAGS` should only
    # include non-library linker flags such as `-L`.
    "ld_flags" : "",

    # We enforce the GNU makefile implicit rule that `LDLIBS` should only
    # include library linker flags/names such as `-lfoo`.
    "ld_libs" : "",
  }

  Checkers = {
    ("verbose", "vl_enable_assert", "vl_line_trace", "vl_W_lint", "vl_W_style",
     "vl_W_fatal", "vl_trace", "vl_coverage", "vl_line_coverage", "vl_toggle_coverage"):
      Checker( lambda v: isinstance(v, bool), "expects a boolean" ),

    ("c_flags", "ld_flags", "ld_libs", "vl_trace_filename"):
      Checker( lambda v: isinstance(v, str),  "expects a string" ),

    ("vl_opt_level", "vl_unroll_count", "vl_unroll_stmts"):
      Checker( lambda v: isinstance(v, int) and v >= 0, "expects an integer >= 0" ),

    "vl_Wno_list": Checker( lambda v: isinstance(v, list) and all(w in VerilogPlaceholderConfigs.Warnings for w in v),
                            "expects a list of warnings" ),

    "vl_xinit": Checker( lambda v: (v in ['zeros', 'ones', 'rand']) or isinstance(v, int),
                  "vl_xinit should be an integer or one of ['zeros', 'ones', 'rand']" ),

    "vl_trace_timescale": Checker( lambda v: isinstance(v, str) and len(v) > 2 and v[-1] == 's' and \
                                    v[-2] in ['p', 'n', 'u', 'm'] and \
                                    all(c.isdigit() for c in v[:-2]),
                                    "expects a timescale string" ),

    "vl_trace_cycle_time": Checker( lambda v: isinstance(v, int) and (v % 2) == 0,
                                    "expects an integer `n` such that `n`*`vl_trace_timescale` is the cycle time" ),

    "vl_mk_dir": Checker( lambda v: isinstance(v, str), "expects a path to directory" ),

    "c_include_path": Checker( lambda v: isinstance(v, list) and all(os.path.isdir(expand(p)) for p in v),
                                "expects a list of paths to directories" ),

    "c_srcs": Checker( lambda v: isinstance(v, list) and all(os.path.isfile(expand(p)) for p in v),
                       "expects a list of paths to files" )
  }

  Warnings = [
    'ALWCOMBORDER', 'ASSIGNIN', 'ASSIGNDLY', 'BLKANDNBLK', 'BLKSEQ',
    'BLKLOOPINIT', 'BSSPACE', 'CASEINCOMPLETE', 'CASEOVERLAP',
    'CASEX', 'CASEWITHX', 'CDCRSTLOGIC', 'CLKDATA', 'CMPCONST',
    'COLONPLUS', 'COMBDLY', 'CONTASSREG', 'DECLFILENAME', 'DEFPARAM',
    'DETECTARRAY', 'ENDLABEL', 'GENCLK', 'IFDEPTH', 'IGNOREDRETURN',
    'IMPERFECTSCH', 'IMPLICIT', 'IMPORTSTAR', 'IMPURE', 'INCABSPATH',
    'INFINITELOOP', 'INITIALDLY', 'LITENDIAN', 'MODDUP', 'MULTIDRIVEN',
    'MULTITOP', 'PINCONNECTEMPTY', 'PINMISSING', 'PINNOCONNECT',
    'PROCASSWIRE', 'REALCVT', 'REDEFMACRO', 'SELRANGE', 'STMTDLY',
    'SYMRSVDWORD', 'SYNCASYNCNET', 'TASKNSVAR', 'TICKCOUNT', 'UNDRIVEN',
    'UNOPT', 'UNOPTFLAT', 'UNOPTTHREADS', 'UNPACKED', 'UNSIGNED', 'UNUSED',
    'USERINFO', 'USERWARN', 'USERERROR', 'USERFATAL', 'VARHIDDEN', 'WIDTH',
    'WIDTHCONCAT',
  ]

  PassName = 'VerilatorImportConfigs'

  #---------------------------------------------------
  # Public APIs
  #---------------------------------------------------

  def setup_configs( s, m, m_tr_namespace ):
    # VerilatorImportConfigs alone does not have the complete information about
    # the module to be imported. For example, we need to read from the placeholder
    # configuration to figure out the pickled file name and the top module name.
    # This method is meant to be called before calling other public APIs.

    s.translated_top_module = m_tr_namespace.translated_top_module
    s.translated_source_file = m_tr_namespace.translated_filename
    s.v_include = m.config_placeholder.v_include
    s.v_libs = m.config_placeholder.v_libs
    # s.src_file = m.config_placeholder.src_file
    s.port_map = m.config_placeholder.port_map
    s.params = m.config_placeholder.params

    if not s.vl_mk_dir:
      s.vl_mk_dir = f'obj_dir_{s.translated_top_module}'

  def get_vl_xinit_value( s ):
    if s.vl_xinit == 'zeros':
      return 0
    elif s.vl_xinit == 'ones':
      return 1
    elif ( s.vl_xinit == 'rand' ) or isinstance( s.vl_xinit, int ):
      return 2
    else:
      raise InvalidPassOptionValue("vl_xinit should be an int or one of 'zeros', 'ones', or 'rand'!")

  def get_vl_xinit_seed( s ):
    if isinstance( s.vl_xinit, int ):
      return s.vl_xinit
    else:
      return 0

  def get_c_wrapper_path( s ):
    return f'{s.translated_top_module}_v.cpp'

  def get_py_wrapper_path( s ):
    return f'{s.translated_top_module}_v.py'

  def get_shared_lib_path( s ):
    return f'lib{s.translated_top_module}_v.so'

  #---------------------
  # Command generation
  #---------------------

  def create_vl_cmd( s ):
    top_module  = f"--top-module {s.translated_top_module}"
    src         = s.translated_source_file
    mk_dir      = f"--Mdir {s.vl_mk_dir}"
    # flist       = "" if s.is_default("v_flist") else \
    #               f"-f {s.v_flist}"
    # We don't need this since all v_libs appear in the translation result
    # vlibs       = "" if not s.v_libs else \
    #               " ".join([f"-v {lib}" for lib in s.v_libs])
    vlibs       = ""
    include     = "" if not s.v_include else \
                  " ".join("-I" + path for path in s.v_include)
    en_assert   = "--assert" if s.vl_enable_assert else ""
    opt_level   = "-O3" if s.is_default( 'vl_opt_level' ) else "-O0"
    loop_unroll = "" if s.vl_unroll_count == 0 else \
                  f"--unroll-count {s.vl_unroll_count}"
    stmt_unroll = "" if s.vl_unroll_stmts == 0 else \
                  f"--unroll-stmts {s.vl_unroll_stmts}"
    trace       = "--trace" if s.vl_trace else ""
    coverage    = "--coverage" if s.vl_coverage else ""
    line_cov    = "--coverage-line" if s.vl_line_coverage else ""
    toggle_cov  = "--coverage-toggle" if s.vl_toggle_coverage else ""
    warnings    = s._create_vl_warning_cmd()

    all_opts = [
      top_module, mk_dir, include, en_assert, opt_level, loop_unroll,
      # stmt_unroll, trace, warnings, flist, src, coverage,
      stmt_unroll, trace, warnings, src, vlibs, coverage,
      line_cov, toggle_cov,
    ]

    return f"verilator --cc {' '.join(opt for opt in all_opts if opt)}"

  def create_cc_cmd( s ):
    c_flags = "-O0 -fPIC -fno-gnu-unique -shared" + \
             ("" if s.is_default("c_flags") else f" {expand(s.c_flags)}")
    c_include_path = " ".join("-I"+p for p in s._get_all_includes() if p)
    out_file = s.get_shared_lib_path()
    c_src_files = " ".join(s._get_c_src_files())
    ld_flags = expand(s.ld_flags)
    ld_libs = s.ld_libs
    coverage = "-DVM_COVERAGE" if s.vl_coverage or \
                                  s.vl_line_coverage or \
                                  s.vl_toggle_coverage else ""
    return f"g++ {c_flags} {c_include_path} {ld_flags}"\
           f" -o {out_file} {c_src_files} {ld_libs} {coverage}"

  def vprint( s, msg, nspaces = 0, use_fill = False ):
    if s.verbose:
      if use_fill:
        print(indent(fill(msg), " "*nspaces))
      else:
        print(indent(msg, " "*nspaces))

  #---------------------
  # Internal helpers
  #---------------------

  def _create_vl_warning_cmd( s ):
    lint = "" if s.is_default("vl_W_lint") else "--Wno-lint"
    style = "" if s.is_default("vl_W_style") else "--Wno-style"
    fatal = "" if s.is_default("vl_W_fatal") else "--Wno-fatal"
    wno = " ".join(f"--Wno-{w}" for w in s.vl_Wno_list)
    return " ".join(w for w in [lint, style, fatal, wno] if w)

  def _get_all_includes( s ):
    includes = s.c_include_path

    # Try to obtain verilator include path either from environment variable
    # or from `pkg-config`
    vl_include_dir = os.environ.get("PYMTL_VERILATOR_INCLUDE_DIR")
    if vl_include_dir is None:
      get_dir_cmd = ["pkg-config", "--variable=includedir", "verilator"]
      try:
        vl_include_dir = \
            subprocess.check_output(get_dir_cmd, stderr = subprocess.STDOUT).strip()
        vl_include_dir = vl_include_dir.decode('ascii')
      except OSError as e:
        vl_include_dir_msg = \
"""\
Cannot locate the include directory of verilator. Please make sure either \
$PYMTL_VERILATOR_INCLUDE_DIR is set or `pkg-config` has been configured properly!
"""
        raise OSError(fill(vl_include_dir_msg)) from e

    # Add verilator include path
    s.vl_include_dir = vl_include_dir
    includes += [vl_include_dir, vl_include_dir + "/vltstd"]

    return includes

  def _get_c_src_files( s ):
    srcs = s.c_srcs
    top_module = s.translated_top_module
    vl_mk_dir = s.vl_mk_dir
    vl_class_mk = f"{vl_mk_dir}/V{top_module}_classes.mk"

    # Add C wrapper
    srcs.append(s.get_c_wrapper_path())

    # Add files listed in class makefile
    with open(vl_class_mk) as class_mk:
      srcs += s._get_srcs_from_vl_class_mk(
          class_mk, vl_mk_dir, "VM_CLASSES_FAST")
      srcs += s._get_srcs_from_vl_class_mk(
          class_mk, vl_mk_dir, "VM_CLASSES_SLOW")
      srcs += s._get_srcs_from_vl_class_mk(
          class_mk, vl_mk_dir, "VM_SUPPORT_FAST")
      srcs += s._get_srcs_from_vl_class_mk(
          class_mk, vl_mk_dir, "VM_SUPPORT_SLOW")
      srcs += s._get_srcs_from_vl_class_mk(
          class_mk, s.vl_include_dir, "VM_GLOBAL_FAST")
      srcs += s._get_srcs_from_vl_class_mk(
          class_mk, s.vl_include_dir, "VM_GLOBAL_SLOW")

    return srcs

  def _get_srcs_from_vl_class_mk( s, mk, path, label ):
    """Return all files under `path` directory in `label` section of `mk`."""
    srcs, found = [], False
    mk.seek(0)
    for line in mk:
      if line.startswith(label):
        found = True
      elif found:
        if line.strip() == "":
          found = False
        else:
          file_name = line.strip()[:-2]
          srcs.append( path + "/" + file_name + ".cpp" )
    return srcs



from .backends.verilog.import_.VerilatorImportConfigs import VerilatorImportConfigs
from .backends.verilog.translation.TranslationConfigs import TranslationConfigs
from .PassGroups import *
from .tracing.TracingConfigs import TracingConfigs



#=========================================================================
# VerilogPlaceholderPass.py
#=========================================================================
# For each placeholder in the component hierarchy, set up default values,
# check if all configs are valid, and pickle the specified Verilog
# source files.
#
# Author : Peitian Pan
# Date   : Jan 27, 2020

import os
import re
import sys
from textwrap import dedent

from pymtl3 import Placeholder
from pymtl3.passes.backends.verilog.util.utility import (
    gen_mapped_ports,
    get_component_unique_name,
)
from pymtl3.passes.backends.verilog.VerilogPlaceholderConfigs import (
    VerilogPlaceholderConfigs,
)
from pymtl3.passes.errors import InvalidPassOptionValue
from pymtl3.passes.PlaceholderConfigs import expand
from pymtl3.passes.PlaceholderPass import PlaceholderPass
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRGetter
from pymtl3.passes.rtlir import RTLIRType as rt


class VerilogPlaceholderPass( PlaceholderPass ):

  def visit_placeholder( s, m ):
    irepr = RTLIRGetter(cache=False).get_component_ifc_rtlir( m )

    s.setup_default_configs( m, irepr )

    # after the previous setup step, placeholder `m` is guaranteed to have
    # config_placeholder attribute.
    if m.config_placeholder.is_valid:
      s.check_valid( m, m.config_placeholder, irepr )
      s.pickle( m, m.config_placeholder, irepr )

  def setup_default_configs( s, m, irepr ):
    if not hasattr( m, 'config_placeholder' ):
      m.config_placeholder = VerilogPlaceholderConfigs()

    cfg = m.config_placeholder

    if cfg.is_valid:
      # If top_module is unspecified, infer it from the component and its
      # parameters. Note we need to make sure the infered top_module matches
      # the default translation result.
      if not cfg.top_module:
        cfg.top_module = irepr.get_name()
        cfg.unique_top_module = get_component_unique_name( irepr )
        # if m.get_parent_object() is None:
        #   # This is a top level module -- use the module name without the
        #   # parameters because we will generate a different wrapper for top!
        #   cfg.top_module = irepr.get_name()
        # else:
        #   # Otherwise use the full name
        #   cfg.top_module = get_component_unique_name( irepr )
      else:
        cfg.unique_top_module = cfg.top_module

      # Only try to infer the name of Verilog source file if both
      # flist and the source file are not specified.
      if not cfg.src_file and not cfg.v_flist:
        cfg.src_file = f"{cfg.top_module}.v"

  def check_valid( s, m, cfg, irepr ):
    pmap, src, flist, include = \
        cfg.port_map, cfg.src_file, cfg.v_flist, cfg.v_include

    # Check params
    for param_name, value in cfg.params.items():
      if not isinstance( value, int ):
        raise InvalidPassOptionValue("params", cfg.params, cfg.PassName,
            f"non-integer parameter {param_name} is not supported yet!")

    # Check port map

    # TODO: this should be based on RTLIR
    # unmapped_unpacked_ports = gen_unpacked_ports( irepr )
    # unmapped_port_names = list(map(lambda x: x[0], list(unmapped_unpacked_ports)))
    # for name in pmap.keys():
    #   if name not in unmapped_port_names:
    #     raise InvalidPassOptionValue("port_map", pmap, cfg.PassName,
    #       f"Port {name} does not exist in component {irepr.get_name()}!")

    for name in pmap.keys():
      try:
        eval(f'm.{name}')
      except:
        raise InvalidPassOptionValue("port_map", pmap, cfg.PassName,
          f"Port {name} does not exist in component {irepr.get_name()}!")

    # Check src_file
    if cfg.src_file and not os.path.isfile(expand(cfg.src_file)):
      raise InvalidPassOptionValue("src_file", cfg.src_file, cfg.PassName,
          'src_file should be a file path!')

    if cfg.v_flist:
      raise InvalidPassOptionValue("v_flist", cfg.v_flist, cfg.PassName,
          'Placeholders backed by Verilog flist are not supported yet!')

    # Check v_flist
    if cfg.v_flist and not os.path.isfile(expand(cfg.v_flist)):
      raise InvalidPassOptionValue("v_flist", cfg.v_flist, cfg.PassName,
          'v_flist should be a file path!')

    # Check v_include
    if cfg.v_include:
      for include in cfg.v_include:
        if not os.path.isdir(expand(include)):
          raise InvalidPassOptionValue("v_include", cfg.v_include, cfg.PassName,
              'v_include should be an array of dir paths!')

  def pickle( s, m, cfg, irepr ):
    # In the namespace cfg:
    #   pickled_dependency_file: path to the pickled Verilog dependency file
    #   pickled_wrapper_file: path to the pickled Verilog wrapper file
    #   pickled_top_module:  name of the top module in the pickled Verilog

    pickled_dependency_file = f'{cfg.unique_top_module}_pickled_dependency.v'
    pickled_wrapper_file = f'{cfg.unique_top_module}_pickled_wrapper.v'
    pickled_top_module = f'{cfg.unique_top_module}_wrapper'

    orig_comp_name       = get_component_unique_name( irepr )
    pickle_dependency    = s._get_v_lib_files( m, cfg, irepr )
    pickle_dependency   += s._get_dependent_verilog_modules( m, cfg, irepr )
    pickle_wrapper, tplt = s._gen_verilog_wrapper( m, cfg, irepr, pickled_top_module )
    def_symbol           = pickled_top_module.upper()

    # The directives are there to prevent potential duplicated definitions
    # pickle_template = dedent(
    #     '''\
    #         // This is a wrapper module that wraps PyMTL placeholder {orig_comp_name}
    #         // This file was generated by PyMTL VerilogPlaceholderPass
    #         `ifndef {def_symbol}
    #         `define {def_symbol}

    #         {pickle_dependency}
    #         {pickle_wrapper}

    #         `endif /* {def_symbol} */
    #     '''
    # )

    cfg.pickled_dependency_file  = pickled_dependency_file
    cfg.pickled_wrapper_file     = pickled_wrapper_file
    cfg.pickled_top_module       = pickled_top_module
    cfg.pickled_wrapper_template = tplt
    cfg.pickled_orig_file        = cfg.src_file
    cfg.def_symbol = def_symbol
    cfg.orig_comp_name = orig_comp_name
    cfg.pickle_dependency = pickle_dependency

    with open( pickled_dependency_file, 'w' ) as fd:
      fd.write( pickle_dependency )

    with open( pickled_wrapper_file, 'w' ) as fd:
      fd.write( pickle_wrapper )

  def _get_v_lib_files( s, m, cfg, irepr ):
    orig_comp_name = get_component_unique_name( irepr )
    tplt = dedent(
        """\
            // The source code below are included because they are specified
            // as the v_lib Verilog placeholder option of component {orig_comp_name}.

            // If you get a duplicated def error from files included below, please
            // make sure they are included either through the v_lib option or the
            // explicit `include statement in the Verilog source code -- if they
            // appear in both then they will be included twice!

            {v_libs}
            // End of all v_lib files for component {orig_comp_name}

        """
    )
    per_file_tplt = dedent(
        """\
            // File {filename} from v_lib option of component {orig_comp_name}
            {file_content}
        """
    )
    v_libs = []

    for filename in cfg.v_libs:
      with open(filename) as fd:
        file_content = fd.read()
        v_libs.append(per_file_tplt.format(**locals()))

    v_libs = ''.join(v_libs)
    return tplt.format(**locals())

  def _get_dependent_verilog_modules( s, m, cfg, irepr ):
    return s._import_sources( cfg, [cfg.src_file] )

  def _gen_verilog_wrapper( s, m, cfg, irepr, pickled_top_module ):
    rtlir_ports = gen_mapped_ports( m, cfg.port_map, cfg.has_clk, cfg.has_reset )

    all_port_names = list(map(lambda x: x[1], rtlir_ports))

    if not cfg.params:
      parameters = irepr.get_params()
    else:
      parameters = cfg.params.items()

    # Port definitions of wrapper
    ports = []
    for idx, (_, name, p, _) in enumerate(rtlir_ports):
      if name:
        if isinstance(p, rt.Array):
          n_dim = p.get_dim_sizes()
          s_dim = ''.join([f'[0:{idx-1}]' for idx in n_dim])
          p = p.get_next_dim_type()
        else:
          s_dim = ''
        ports.append(
            f"  {p.get_direction()} logic [{p.get_dtype().get_length()}-1:0]"\
            f" {name} {s_dim}{'' if idx == len(rtlir_ports)-1 else ','}"
        )

    # The wrapper has to have an unused clk port to make verilator
    # VCD tracing work.
    if 'clk' not in all_port_names:
      ports.insert( 0, '  input logic clk,' )

    if 'reset' not in all_port_names:
      ports.insert( 0, '  input logic reset,' )

    # Parameters passed to the module to be parametrized
    params = [
      f"    .{param}( {val} ){'' if idx == len(parameters)-1 else ','}"\
      for idx, (param, val) in enumerate(parameters)
    ]

    # Connections between top module and inner module
    connect_ports = [
      f"    .{name}( {name} ){'' if idx == len(rtlir_ports)-1 else ','}"\
      for idx, (_, name, p, _) in enumerate(rtlir_ports) if name
    ]

    lines = [
      f"module {pickled_top_module}",
      "(",
    ] + ports + [
      ");",
      f"  {cfg.top_module}",
      "  #(",
    ] + params + [
      "  ) v",
      "  (",
    ] + connect_ports + [
      "  );",
      "endmodule",
    ]

    template_lines = [
      "module {top_module_name}",
      "(",
    ] + ports + [
      ");",
      f"  {cfg.top_module}",
      "  #(",
    ] + params + [
      "  ) v",
      "  (",
    ] + connect_ports + [
      "  );",
      "endmodule",
    ]

    return '\n'.join( line for line in lines ), '\n'.join( line for line in template_lines )

  #-----------------------------------------------------------------------
  # import_sources
  #-----------------------------------------------------------------------
  # The right way to do this is to use a recursive function like I have
  # done below. This ensures that files are inserted into the output stream
  # in the correct order. -cbatten

  # Regex to extract verilog filenames from `include statements

  _include_re = re.compile( r'"(?P<filename>[\w/\.-]*)"' )

  def _output_verilog_file( s, include_path, verilog_file ):
    code = ""
    with open(verilog_file) as fp:

      short_verilog_file = verilog_file
      if verilog_file.startswith( include_path+"/" ):
        short_verilog_file = verilog_file[len(include_path+"/"):]

      code += '`line 1 "{}" 0\n'.format( short_verilog_file )

      line_num = 0
      for line in fp:
        line_num += 1
        if '`include' in line:
          filename = s._include_re.search( line ).group( 'filename' )
          fullname = os.path.join( include_path, filename )
          code += s._output_verilog_file( include_path, fullname )
          code += '\n'
          code += '`line {} "{}" 0\n'.format( line_num+1, short_verilog_file )
        else:
          code += line
    return code

  def _import_sources( s, cfg, source_list ):
    """Import Verilog source from all Verilog files source_list, as well
    as any source files specified by `include within those files.
    """

    code = ""

    if not source_list:
      return

    # We will use the first verilog file to find the root of PyMTL project

    first_verilog_file = source_list[0]

    # All verilog includes are relative to the root of the PyMTL project.
    # We identify the root of the PyMTL project by looking for the special
    # .pymtl_sim_root file.

    _path = os.path.dirname( first_verilog_file )
    special_file_found = False
    include_path = os.path.dirname( os.path.abspath( first_verilog_file ) )
    while include_path != "/":
      if os.path.exists( include_path + os.path.sep + ".pymtl_sim_root" ):
        special_file_found = True
        sys.path.insert(0,include_path)
        break
      include_path = os.path.dirname(include_path)

    # Append the user-defined include path to include_path
    # NOTE: the current pickler only supports one include path. If v_include
    # config is present, use it instead.

    if cfg.v_include:
      if len(cfg.v_include) != 1:
        raise InvalidPassOptionValue("v_include", cfg.v_include, cfg.PassName,
            'the current pickler only supports one user-defined v_include path...')
      include_path = cfg.v_include[0]

    # If we could not find the special .pymtl-python-path file, then assume
    # the include directory is the same as the directory that contains the
    # first verilog file.

    if not special_file_found and not cfg.v_include:
      include_path = os.path.dirname( os.path.abspath( first_verilog_file ) )

    # Regex to extract verilog filenames from `include statements

    s._include_re = re.compile( r'"(?P<filename>[\w/\.-]*)"' )

    # Iterate through all source files and add any `include files to the
    # list of source files to import.

    for source in source_list:
      code += s._output_verilog_file( include_path, source )

    return code



#=========================================================================
# BehavioralRTLIRTypeCheckL3Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 30, 2019
"""Provide L3 behavioral RTLIR type check pass."""

from pymtl3.passes.rtlir.errors import PyMTLTypeError
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.rtlir.rtype import RTLIRType as rt

from .BehavioralRTLIRTypeCheckL2Pass import (
    BehavioralRTLIRTypeCheckL2Pass,
    BehavioralRTLIRTypeCheckVisitorL2,
)


class BehavioralRTLIRTypeCheckL3Pass( BehavioralRTLIRTypeCheckL2Pass ):
  def get_visitor_class( s ):
    return BehavioralRTLIRTypeCheckVisitorL3

class BehavioralRTLIRTypeCheckVisitorL3( BehavioralRTLIRTypeCheckVisitorL2 ):
  def __init__( s, component, freevars, accessed, tmpvars, rtlir_getter ):
    super().__init__( component, freevars, accessed, tmpvars, rtlir_getter )
    s.type_expect[ 'Attribute' ] = (
      ( 'value', (rt.Component, rt.Signal),
                 'the base of an attribute must be one of: component, signal!' ),
    )

  def visit_Attribute( s, node ):
    if isinstance( node.value.Type, rt.Signal ):
      dtype = node.value.Type.get_dtype()
      if not isinstance( dtype, rdt.Struct ):
        raise PyMTLTypeError( s.blk, node.ast,
          'attribute base should be a struct signal!'
        )
      if not dtype.has_property( node.attr ):
        raise PyMTLTypeError( s.blk, node.ast,
          f'{dtype.get_name()} does not have field {node.attr}!' )
      dtype = dtype.get_property( node.attr )
      if isinstance( node.value.Type, rt.Port ):
        rtype = rt.Port( node.value.Type.get_direction(), dtype )
      elif isinstance( node.value.Type, rt.Wire ):
        rtype = rt.Wire( dtype )
      elif isinstance( node.value.Type, rt.Const ):
        obj = node.value.Type.get_object()
        if obj is None:
          rtype = rt.Const( dtype )
        else:
          try:
            rtype = rt.Const( dtype, getattr( obj, node.attr ) )
          except AttributeError:
            rtype = rt.Const( dtype )
      else:
        raise PyMTLTypeError( s.blk, node.ast,
          f'unrecognized signal type {node.value.Type}!' )
      node.Type = rtype

    else:
      super().visit_Attribute( node )

  def visit_StructInst( s, node ):
    cls = node.struct

    dtype = rdt.get_rtlir_dtype( cls() )
    all_properties = dtype.get_all_properties()

    if len( all_properties ) != len( node.values ):
      raise PyMTLTypeError( s.blk, node.ast,
        f"BitStruct {cls.__name__} has {len(all_properties)} fields but only {len(node.values)} arguments are given!" )

    all_types = zip( node.values, list(all_properties.items()) )

    for idx, ( value, ( name, field ) ) in enumerate( all_types ):
      s.visit( value )
      # Expect each argument to be a signal
      if not isinstance( value.Type, rt.Signal ):
        raise PyMTLTypeError( s.blk, node.ast,
          f"argument #{idx} has type {value.Type} but not a signal!" )
      v_dtype = value.Type.get_dtype()
      # Expect each argument to have data type which corresponds to the field
      if v_dtype != field:
        raise PyMTLTypeError( s.blk, node.ast,
          f"Expected argument#{idx} ( field {name} ) to have type {field}, but got {v_dtype}." )

    node.Type = rt.Const( dtype )



from .import_.VerilatorImportConfigs import VerilatorImportConfigs
from .import_.VerilatorImportPass import VerilatorImportPass
from .tbgen.VerilogTBGenPass import VerilogTBGenPass
from .translation.TranslationConfigs import TranslationConfigs
from .translation.TranslationPass import TranslationPass
from .TranslationImportPass import TranslationImportPass
from .VerilogPlaceholderConfigs import VerilogPlaceholderConfigs
from .VerilogPlaceholderPass import VerilogPlaceholderPass



#=========================================================================
# ImportPass.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 25, 2019
"""Provide a pass that imports arbitrary SystemVerilog modules."""

import copy
import importlib
import json
import linecache
import os
import shutil
import subprocess
import sys
from importlib import reload
from itertools import cycle
from textwrap import indent

from pymtl3 import Placeholder
from pymtl3.datatypes import Bits, is_bitstruct_class, is_bitstruct_inst, mk_bits
from pymtl3.dsl import Component
from pymtl3.passes.BasePass import BasePass
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRGetter
from pymtl3.passes.rtlir import RTLIRType as rt

from ..errors import VerilogImportError
from ..util.utility import (
    expand,
    gen_mapped_ports,
    get_component_unique_name,
    get_rtype,
    make_indent,
    wrap,
)
from .VerilatorImportConfigs import VerilatorImportConfigs


class VerilatorImportPass( BasePass ):
  """Import an arbitrary SystemVerilog module as a PyMTL component.

  The import pass takes as input a PyMTL component hierarchy where
  the components to be imported have an `import` entry in their parameter
  dictionary( set by calling the `set_parameter` PyMTL API ).
  This pass assumes the modules to be imported are located in the current
  directory and have name {full_name}.v where `full_name` is the name of
  component's class concatanated with the list of arguments of its construct
  method. It has the following format:
      > {module_name}[__{parameter_name}_{parameter_value}]*
  As an example, component mux created through `mux = Mux(Bits32, 2)` has a
  full name `Mux__Type_Bits32__ninputs_2`.
  The top module inside the target .v file should also have a full name.
  """
  def __call__( s, top ):
    s.top = top
    if not top._dsl.constructed:
      raise VerilogImportError( top,
        f"please elaborate design {top} before applying the import pass!" )
    ret = s.traverse_hierarchy( top )
    if ret is None:
      ret = top
    else:
      ret.elaborate()
    return ret

  def traverse_hierarchy( s, m ):
    # Import can only be performed on Placeholders
    if hasattr(m, 'config_placeholder') and m.config_placeholder.is_valid:
      if not hasattr(m, s.get_config_name()):
        setattr(m, s.get_config_name(), VerilatorImportConfigs())
      return s.do_import( m )

    else:
      for child in m.get_child_components(repr):
        s.traverse_hierarchy( child )

  def do_import( s, m ):
    try:
      imp = s.get_imported_object( m )
      if m is s.top:
        return imp
      else:
        s.top.replace_component_with_obj( m, imp )
    except AssertionError as e:
      msg = '' if e.args[0] is None else e.args[0]
      raise VerilogImportError( m, msg )

  #-----------------------------------------------------------------------
  # Backend-specific methods
  #-----------------------------------------------------------------------

  def get_backend_name( s ):
    return "verilog"

  def get_config_name( s ):
    return "config_verilog_import"

  def get_placeholder_config( s, m ):
    return m.config_placeholder

  def get_config( s, m ):
    return m.config_verilog_import

  def get_translation_namespace( s, m ):
    return m._pass_verilog_translation

  def get_gen_mapped_port( s ):
    return gen_mapped_ports

  #-----------------------------------------------------------------------
  # is_cached
  #-----------------------------------------------------------------------

  def is_cached( s, m, ip_cfg ):
    # Only components translated by pymtl translation passes will be cached
    try:
      is_same = s.get_translation_namespace(m).is_same
    except AttributeError:
      is_same = False

    # Check if the verilated model is cached
    is_source_cached = False
    obj_dir = ip_cfg.vl_mk_dir
    c_wrapper = ip_cfg.get_c_wrapper_path()
    py_wrapper = ip_cfg.get_py_wrapper_path()
    shared_lib = ip_cfg.get_shared_lib_path()
    if is_same and os.path.exists(obj_dir) and os.path.exists(c_wrapper) and \
       os.path.exists(py_wrapper) and os.path.exists(shared_lib):
      is_source_cached = True

    # Check if the configurations from the last run are the same
    is_config_cached = False
    config_file = f'pymtl_import_config_{ip_cfg.translated_top_module}.json'
    new_cfg = s.serialize_cfg( ip_cfg )
    if os.path.exists(config_file):
      with open(config_file) as fd:
        if s.is_same_cfg( json.load( fd ), new_cfg ):
          is_config_cached = True

    return is_source_cached and is_config_cached, config_file, new_cfg

  #-----------------------------------------------------------------------
  # get_imported_object
  #-----------------------------------------------------------------------

  def get_imported_object( s, m ):
    ph_cfg = s.get_placeholder_config( m )
    ip_cfg = s.get_config( m )
    ip_cfg.setup_configs( m, s.get_translation_namespace(m) )

    rtype = RTLIRGetter(cache=False).get_component_ifc_rtlir( m )

    # Now we selectively unpack array of ports if they are referred to in
    # port_map
    ports = s.get_gen_mapped_port()( m, ph_cfg.port_map, ph_cfg.has_clk, ph_cfg.has_reset )

    cached, config_file, cfg_d = s.is_cached( m, ip_cfg )

    s.create_verilator_model( m, ph_cfg, ip_cfg, cached )

    port_cdefs = s.create_verilator_c_wrapper( m, ph_cfg, ip_cfg, ports, cached )

    s.create_shared_lib( m, ph_cfg, ip_cfg, cached )

    symbols = s.create_py_wrapper( m, ph_cfg, ip_cfg, rtype, ports, port_cdefs, cached )

    imp = s.import_component( m, ph_cfg, ip_cfg, symbols )

    imp._ip_cfg = ip_cfg
    imp._ph_cfg = ph_cfg
    imp._ports = ports

    # Dump configuration dict to config_file
    with open( config_file, 'w' ) as fd:
      json.dump( cfg_d, fd, indent = 4 )

    return imp

  #-----------------------------------------------------------------------
  # create_verilator_model
  #-----------------------------------------------------------------------

  def create_verilator_model( s, m, ph_cfg, ip_cfg, cached ):
    """Verilate module `m`."""
    ip_cfg.vprint("\n=====Verilate model=====")

    if not cached:
      # Generate verilator command
      cmd = ip_cfg.create_vl_cmd()

      # Remove obj_dir directory if it already exists.
      # obj_dir is where the verilator output ( C headers and sources ) is stored
      obj_dir = ip_cfg.vl_mk_dir
      if os.path.exists( obj_dir ):
        shutil.rmtree( obj_dir )

      succeeds = True

      # Try to call verilator
      try:
        ip_cfg.vprint(f"Verilating {ip_cfg.translated_top_module} with command:", 2)
        ip_cfg.vprint(f"{cmd}", 4)
        subprocess.check_output(
            cmd, stderr = subprocess.STDOUT, shell = True )
      except subprocess.CalledProcessError as e:
        succeeds = False
        err_msg = e.output if not isinstance(e.output, bytes) else \
                  e.output.decode('utf-8')
        import_err_msg = \
            f"Fail to verilate model {ip_cfg.translated_top_module}\n"\
            f"  Verilator command:\n{indent(cmd, '  ')}\n\n"\
            f"  Verilator output:\n{indent(wrap(err_msg), '  ')}\n"

      if not succeeds:
        raise VerilogImportError(m, import_err_msg)

      ip_cfg.vprint("Successfully verilated the given model!", 2)

    else:
      ip_cfg.vprint(f"{ip_cfg.translated_top_module} not verilated because it's cached!", 2)

  #-----------------------------------------------------------------------
  # create_verilator_c_wrapper
  #-----------------------------------------------------------------------

  def create_verilator_c_wrapper(
      s, m, ph_cfg, ip_cfg, ports, cached ):
    """Return the file name of generated C component wrapper.

    Create a C wrapper that calls verilator C API and provides interfaces
    that can be later called through CFFI.
    """
    component_name = ip_cfg.translated_top_module
    dump_vcd = int(ip_cfg.vl_trace)
    vcd_timescale = ip_cfg.vl_trace_timescale
    half_cycle_time = ip_cfg.vl_trace_cycle_time // 2
    external_trace = int(ip_cfg.vl_line_trace)
    wrapper_name = ip_cfg.get_c_wrapper_path()
    verilator_xinit_value = ip_cfg.get_vl_xinit_value()
    verilator_xinit_seed = ip_cfg.get_vl_xinit_seed()
    ip_cfg.vprint("\n=====Generate C wrapper=====")

    # The wrapper template should be in the same directory as this file
    template_name = \
      os.path.dirname( os.path.abspath( __file__ ) ) + \
      os.path.sep + 'verilator_wrapper.c.template'

    # Generate port declarations for the verilated model in C
    port_defs = []
    for _, v_name, port, _ in ports:
      if v_name:
        port_defs.append( s.gen_signal_decl_c( v_name, port ) )
    port_cdefs = copy.copy( port_defs )
    make_indent( port_defs, 2 )
    port_defs = '\n'.join( port_defs )

    # Generate initialization statements for in/out ports
    port_inits = []
    for _, v_name, port, _ in ports:
      if v_name:
        port_inits.extend( s.gen_signal_init_c( v_name, port ) )
    make_indent( port_inits, 1 )
    port_inits = '\n'.join( port_inits )

    # Fill in the C wrapper template
    with open(template_name) as template:
      with open( wrapper_name, 'w' ) as output:
        c_wrapper = template.read()
        c_wrapper = c_wrapper.format( **locals() )
        output.write( c_wrapper )

    ip_cfg.vprint(f"Successfully generated C wrapper {wrapper_name}!", 2)
    return port_cdefs

  #-----------------------------------------------------------------------
  # create_shared_lib
  #-----------------------------------------------------------------------

  def create_shared_lib( s, m, ph_cfg, ip_cfg, cached ):
    """Return the name of compiled shared lib."""
    full_name = ip_cfg.translated_top_module
    dump_vcd = ip_cfg.vl_trace
    ip_cfg.vprint("\n=====Compile shared library=====")

    if not cached:
      cmd = ip_cfg.create_cc_cmd()

      succeeds = True

      # Try to call the C compiler
      try:
        ip_cfg.vprint("Compiling shared library with command:", 2)
        ip_cfg.vprint(f"{cmd}", 4)
        subprocess.check_output(
            cmd,
            stderr = subprocess.STDOUT,
            shell = True,
            universal_newlines = True
        )
      except subprocess.CalledProcessError as e:
        succeeds = False
        err_msg = e.output if not isinstance(e.output, bytes) else \
                  e.output.decode('utf-8')
        import_err_msg = \
            f"Failed to compile Verilated model into a shared library:\n"\
            f"  C compiler command:\n{indent(cmd, '  ')}\n\n"\
            f"  C compiler output:\n{indent(wrap(err_msg), '  ')}\n"

      if not succeeds:
        raise VerilogImportError(m, import_err_msg)

      ip_cfg.vprint(f"Successfully compiled shared library "\
                    f"{ip_cfg.get_shared_lib_path()}!", 2)

    else:
      ip_cfg.vprint("Didn't compile shared library because it's cached!", 2)

  #-----------------------------------------------------------------------
  # create_py_wrapper
  #-----------------------------------------------------------------------

  def create_py_wrapper(
      s, m, ph_cfg, ip_cfg, rtype, ports, port_cdefs, cached ):
    """Return the file name of the generated PyMTL component wrapper."""
    ip_cfg.vprint("\n=====Generate PyMTL wrapper=====")

    # Load the wrapper template
    template_name = \
      os.path.dirname( os.path.abspath( __file__ ) ) + \
      os.path.sep + 'verilator_wrapper.py.template'
    wrapper_name = ip_cfg.get_py_wrapper_path()

    # Port definitions of verilated model
    make_indent( port_cdefs, 4 )

    # Port definition in PyMTL style
    symbols, port_defs = s.gen_signal_decl_py( rtype )
    make_indent( port_defs, 2 )

    # Set upblk inputs and outputs
    set_comb_input, structs_input   = s.gen_comb_input( ports, symbols )
    set_comb_output, structs_output = s.gen_comb_output( ports, symbols )
    make_indent( structs_input, 2 )
    make_indent( structs_output, 2 )
    make_indent( set_comb_input, 3 )
    make_indent( set_comb_output, 3 )

    # Line trace
    line_trace = s.gen_line_trace_py( ports )

    # Internal line trace
    in_line_trace = s.gen_internal_line_trace_py( ports )

    # External trace function definition
    if ip_cfg.vl_line_trace:
      external_trace_c_def = f'void trace( V{ip_cfg.translated_top_module}_t *, char * );'
    else:
      external_trace_c_def = ''

    # Fill in the python wrapper template
    with open(template_name) as template:
      with open( wrapper_name, 'w' ) as output:
        py_wrapper = template.read()
        py_wrapper = py_wrapper.format(
          component_name        = ip_cfg.translated_top_module,
          has_clk               = int(ph_cfg.has_clk),
          clk                   = 'inv_clk' if not ph_cfg.has_clk else \
                                  next(filter(lambda x: x[0][0]=='clk', ports))[1],
          lib_file              = ip_cfg.get_shared_lib_path(),
          port_cdefs            = ('  '*4+'\n').join( port_cdefs ),
          port_defs             = '\n'.join( port_defs ),
          structs_input         = '\n'.join( structs_input ),
          structs_output        = '\n'.join( structs_output ),
          set_comb_input        = '\n'.join( set_comb_input ),
          set_comb_output       = '\n'.join( set_comb_output ),
          line_trace            = line_trace,
          in_line_trace         = in_line_trace,
          dump_vcd              = int(ip_cfg.vl_trace),
          has_vl_trace_filename = bool(ip_cfg.vl_trace_filename),
          vl_trace_filename     = ip_cfg.vl_trace_filename,
          external_trace        = int(ip_cfg.vl_line_trace),
          trace_c_def           = external_trace_c_def,
        )
        output.write( py_wrapper )

    ip_cfg.vprint(f"Successfully generated PyMTL wrapper {wrapper_name}!", 2)
    return symbols

  #-----------------------------------------------------------------------
  # import_component
  #-----------------------------------------------------------------------

  def import_component( s, m, ph_cfg, ip_cfg, symbols ):
    """Return the PyMTL component imported from `wrapper_name`.v."""
    ip_cfg.vprint("=====Create python object=====")

    component_name = ip_cfg.translated_top_module
    # Get the name of the wrapper Python module
    wrapper_name = ip_cfg.get_py_wrapper_path()
    wrapper = wrapper_name.split('.')[0]

    # Add CWD to sys.path so we can import from the current directory
    if not os.getcwd() in sys.path:
      sys.path.append( os.getcwd() )

    # Check linecache in case the wrapper file has been modified
    linecache.checkcache()

    # Import the component from python wrapper

    if wrapper in sys.modules:
      # Reload the wrapper module in case the user has updated the wrapper
      reload(sys.modules[wrapper])
    else:
      # importlib.import_module inserts the wrapper module into sys.modules
      importlib.import_module(wrapper)

    # Try to access the top component class from the wrapper module
    try:
      imp_class = getattr( sys.modules[wrapper], component_name )
    except AttributeError as e:
      raise VerilogImportError(m,
          f"internal error: PyMTL wrapper {wrapper_name} does not have "
          f"top component {component_name}!") from e

    imp = imp_class()
    ip_cfg.vprint(f"Successfully created python object of {component_name}!", 2)

    # Update the global namespace of `construct` so that the struct and interface
    # classes defined previously can still be used in the imported model.
    imp.construct.__globals__.update( symbols )

    ip_cfg.vprint("Import succeeds!")
    return imp

  #-------------------------------------------------------------------------
  # Serialize and compare configurations
  #-------------------------------------------------------------------------

  def serialize_cfg( s, ip_cfg ):
    d = {}
    s._volatile_configs = [
      'vl_line_trace', 'vl_coverage', 'vl_line_coverage', 'vl_toggle_coverage',
      'vl_mk_dir', 'vl_enable_assert', 'vl_opt_level',
      'vl_unroll_count', 'vl_unroll_stmts',
      'vl_W_lint', 'vl_W_style', 'vl_W_fatal', 'vl_Wno_list',
      'vl_xinit', 'vl_trace',
      'vl_trace_timescale', 'vl_trace_cycle_time',
      'c_flags', 'c_include_path', 'c_srcs',
      'ld_flags', 'ld_libs',
    ]
    d['ImportPassName'] = 'VerilatorImportPass'
    for cfg in s._volatile_configs:
      d[cfg] = copy.deepcopy(getattr( ip_cfg, cfg ))

    return d

  def is_same_cfg( s, prev, new ):
    _volatile_configs = copy.copy(s._volatile_configs)
    _volatile_configs.append('ImportPassName')
    return all(prev[cfg] == new[cfg] for cfg in _volatile_configs)

  #-------------------------------------------------------------------------
  # gen_signal_decl_c
  #-------------------------------------------------------------------------

  def gen_signal_decl_c( s, name, port ):
    """Return C variable declaration of `port`."""
    c_dim = s._get_c_dim( port )
    nbits = s._get_c_nbits( port )
    UNSIGNED_8  = 'unsigned char'
    UNSIGNED_16 = 'unsigned short'
    UNSIGNED_32 = 'unsigned int'
    if sys.maxsize > 2**32:
      UNSIGNED_64 = 'unsigned long'
    else:
      UNSIGNED_64 = 'unsigned long long'
    if    nbits <= 8:  data_type = UNSIGNED_8
    elif  nbits <= 16: data_type = UNSIGNED_16
    elif  nbits <= 32: data_type = UNSIGNED_32
    elif  nbits <= 64: data_type = UNSIGNED_64
    else:              data_type = UNSIGNED_32
    name = s._verilator_name( name )
    return f'{data_type} * {name}{c_dim};'

  #-------------------------------------------------------------------------
  # gen_signal_init_c
  #-------------------------------------------------------------------------

  def gen_signal_init_c( s, name, port ):
    """Return C port variable initialization."""
    ret       = []
    c_dim     = s._get_c_dim( port )
    nbits     = s._get_c_nbits( port )
    deference = '&' if nbits <= 64 else ''
    name      = s._verilator_name( name )

    if c_dim:
      n_dim_size = s._get_c_n_dim( port )
      sub = ""
      for_template = \
"""\
for ( int i_{idx} = 0; i_{idx} < {dim_size}; i_{idx}++ )
"""
      assign_template = \
"""\
m->{name}{sub} = {deference}model->{name}{sub};
"""

      for idx, dim_size in enumerate( n_dim_size ):
        ret.append( for_template.format( **locals() ) )
        sub += f"[i_{idx}]"

      ret.append( assign_template.format( **locals() ) )

      # Indent the for loop
      for start, dim_size in enumerate( n_dim_size ):
        for idx in range( start + 1, len( n_dim_size ) + 1 ):
          ret[ idx ] = "  " + ret[ idx ]

    else:
      ret.append(f'm->{name} = {deference}model->{name};')

    return ret

  #-------------------------------------------------------------------------
  # gen_signal_decl_py
  #-------------------------------------------------------------------------

  def gen_signal_decl_py( s, rtype ):
    """Return the PyMTL definition of all interface ports of `rtype`."""

    #-----------------------------------------------------------------------
    # Methods that generate signal declarations
    #-----------------------------------------------------------------------

    def gen_dtype_str( symbols, dtype ):
      if isinstance( dtype, rdt.Vector ):
        nbits = dtype.get_length()
        Bits_name = f"Bits{nbits}"
        if Bits_name not in symbols and nbits >= 256:
          Bits_class = mk_bits( nbits )
          symbols.update( { Bits_name : Bits_class } )
        return f'Bits{dtype.get_length()}'
      elif isinstance( dtype, rdt.Struct ):
        # It is possible to reuse the existing struct class because its __init__
        # can be called without arguments.
        name, cls = dtype.get_name(), dtype.get_class()
        if name not in symbols:
          symbols.update( { name : cls } )
        return name
      else:
        assert False, f"unrecognized data type {dtype}!"

    def gen_port_decl_py( ports ):
      symbols, decls = {}, []
      for id_, _port in ports:
        if id_ not in ['clk', 'reset']:
          if isinstance( _port, rt.Array ):
            n_dim = _port.get_dim_sizes()
            rhs = "{direction}( {dtype} )"
            port = _port.get_sub_type()
            _n_dim = copy.copy( n_dim )
            _n_dim.reverse()
            for length in _n_dim:
              rhs = f"[ {rhs} for _ in range({length}) ]"
          else:
            rhs = "{direction}( {dtype} )"
            port = _port
          direction = s._get_direction( port )
          dtype = gen_dtype_str( symbols, port.get_dtype() )
          rhs = rhs.format( **locals() )
          decls.append(f"s.{id_} = {rhs}")
      return symbols, decls

    def gen_ifc_decl_py( ifcs ):

      def gen_ifc_str( symbols, ifc ):

        def _get_arg_str( name, obj ):
          # Support common python types and Bits/BitStruct
          if isinstance( obj, (int, bool, str) ):
            return str(obj)
          elif obj == None:
            return 'None'
          elif isinstance( obj, Bits ):
            nbits = obj.nbits
            value = int(obj)
            Bits_name = f"Bits{nbits}"
            Bits_arg_str = f"{Bits_name}( {value} )"
            if Bits_name not in symbols and nbits >= 256:
              Bits_class = mk_bits( nbits )
              symbols.update( { Bits_name : Bits_class } )
            return Bits_arg_str
          elif is_bitstruct_inst( obj ):
            raise TypeError("Do you really want to pass in an instance of "
                            "a BitStruct? Contact PyMTL developers!")
            # This is hacky: we don't know how to construct an object that
            # is the same as `obj`, but we do have the object itself. If we
            # add `obj` to the namespace of `construct` everything works fine
            # but the user cannot tell what object is passed to the constructor
            # just from the code.
            # Do not use a double underscore prefix because that will be
            # interpreted differently by the Python interpreter
            # bs_name = ("_" if name[0] != "_" else "") + name + "_obj"
            # if bs_name not in symbols:
              # symbols.update( { bs_name : obj } )
            # return bs_name
          elif isinstance( obj, type ) and issubclass( obj, Bits ):
            nbits = obj.nbits
            Bits_name = f"Bits{nbits}"
            if Bits_name not in symbols and nbits >= 256:
              Bits_class = mk_bits( nbits )
              symbols.update( { Bits_name : Bits_class } )
            return Bits_name
          elif is_bitstruct_class(obj):
            BitStruct_name = obj.__name__
            if BitStruct_name not in symbols:
              symbols.update( { BitStruct_name : obj } )
            return BitStruct_name
          # FIXME formalize this
          elif isinstance( obj, type ) and hasattr( obj, 'req' ) and hasattr( obj, 'resp' ):
            symbols.update( { obj.__name__ : obj } )
            return obj.__name__
          raise TypeError( f"Interface constructor argument {obj} is not an int/Bits/BitStruct/TypeTuple!" )

        name, cls = ifc.get_name(), ifc.get_class()
        if name not in symbols:
          symbols.update( { name : cls } )
        arg_list = []
        args = ifc.get_args()
        for idx, obj in enumerate(args[0]):
          arg_list.append( _get_arg_str( f"_ifc_arg{idx}", obj ) )
        for arg_name, arg_obj in args[1].items():
          arg_list.append( f"{arg_name} = {_get_arg_str( arg_name, arg_obj )}" )
        return name, ', '.join( arg_list )

      symbols, decls = {}, []
      for id_, ifc in ifcs:
        if isinstance( ifc, rt.Array ):
          n_dim = ifc.get_dim_sizes()
          rhs = "{ifc_class}({ifc_params})"
          _ifc = ifc.get_sub_type()
          _n_dim = copy.copy( n_dim )
          _n_dim.reverse()
          for length in _n_dim:
            rhs = f"[ {rhs} for _ in range({length}) ]"
        else:
          rhs = "{ifc_class}({ifc_params})"
          _ifc = ifc
        ifc_class, ifc_params = gen_ifc_str( symbols, _ifc )
        if ifc_params:
          ifc_params = " " + ifc_params + " "
        rhs = rhs.format( **locals() )
        decls.append(f"s.{id_} = {rhs}")
      return symbols, decls

    #-----------------------------------------------------------------------
    # Method gen_signal_decl_py
    #-----------------------------------------------------------------------

    ports = rtype.get_ports_packed()
    ifcs = rtype.get_ifc_views_packed()

    p_symbols, p_decls = gen_port_decl_py( ports )
    i_symbols, i_decls = gen_ifc_decl_py( ifcs )

    return {**p_symbols, **i_symbols}, p_decls + i_decls

  #-----------------------------------------------------------------------
  # Methods that generate python signal writes
  #-----------------------------------------------------------------------

  def _gen_vector_write( s, d, lhs, rhs, dtype, pos ):
    nbits = dtype.get_length()
    l, r = pos, pos+nbits
    if d == 'i':
      ret = [ f"{rhs}[{l}:{r}] = {lhs}" ]
    else:
      ret = [ f"{lhs} = {rhs}[{l}:{r}]" ]
    return ret, r

  def _gen_struct_write( s, d, lhs, rhs, dtype, pos ):
    ret = []
    all_properties = reversed(list(dtype.get_all_properties().items()))
    for name, field in all_properties:
      _ret, pos = s._gen_write_dispatch( d, f"{lhs}.{name}", rhs, field, pos )
      ret.extend( _ret )
    return ret, pos

  def _gen_packed_array_write( s, d, lhs, rhs, dtype, n_dim, pos ):
    if not n_dim:
      return s._gen_write_dispatch( d, lhs, rhs, dtype, pos )
    # Recursively generate array
    ret = []
    for idx in range(n_dim[0]):
      _ret, pos = s._gen_packed_array_write( d, f"{lhs}[{idx}]", rhs, dtype, n_dim[1:], pos )
      ret.extend( _ret )
    return ret, pos

  def _gen_write_dispatch( s, d, lhs, rhs, dtype, pos ):
    if isinstance( dtype, rdt.Vector ):
      return s._gen_vector_write( d, lhs, rhs, dtype, pos )
    elif isinstance( dtype, rdt.Struct ):
      return s._gen_struct_write( d, lhs, rhs, dtype, pos )
    elif isinstance( dtype, rdt.PackedArray ):
      n_dim = dtype.get_dim_sizes()
      sub_dtype = dtype.get_sub_dtype()
      return s._gen_packed_array_write( d, lhs, rhs, sub_dtype, n_dim, pos )
    assert False, f"unrecognized data type {dtype}!"

  #-------------------------------------------------------------------------
  # gen_comb_input
  #-------------------------------------------------------------------------

  def gen_port_vector_input( s, lhs, rhs, mangled_rhs, dtype, symbols ):
    dtype_nbits = dtype.get_length()
    blocks   = [ f's.{mangled_rhs} = Wire( {s._gen_bits_decl(dtype_nbits)} )',
                 '@s.update',
                 f'def isignal_{mangled_rhs}():',
                 f'  s.{mangled_rhs} = {rhs}' ]
    set_comb = ( s._gen_ref_write( lhs, 's.'+mangled_rhs, dtype_nbits ) )
    return set_comb, blocks

  def gen_port_struct_input( s, lhs, rhs, mangled_rhs, dtype, symbols ):
    dtype_nbits = dtype.get_length()
    # If the top-level signal is a struct, we add the datatype to symbol?
    dtype_name = dtype.get_class().__name__
    if dtype_name not in symbols:
      symbols[dtype_name] = dtype.get_class()

    blocks   = [ f's.{mangled_rhs} = Wire( {s._gen_bits_decl(dtype_nbits)} )',
                 '@s.update',
                 f'def istruct_{mangled_rhs}():' ]

    # We write each struct field to tmp
    upblk_content, pos = s._gen_struct_write( 'i', rhs, 's.'+mangled_rhs, dtype, 0 )
    assert pos == dtype_nbits
    make_indent( upblk_content, 1 )
    blocks += upblk_content

    # We don't create a new struct if we are copying values from pymtl
    # land to verilator, i.e. this port is the input to the imported
    # component.
    # At the end, we write tmp to the corresponding CFFI variable
    set_comb = s._gen_ref_write( lhs, 's.'+mangled_rhs, dtype_nbits )
    return set_comb, blocks

  def gen_port_input( s, lhs, rhs, pnames, dtype, symbols ):
    rhs = rhs.format(next(pnames))

    # We always name mangle now
    mangled_rhs = s._pymtl_name_mangle( rhs )

    if isinstance( dtype, rdt.Vector ):
      return s.gen_port_vector_input( lhs, rhs, mangled_rhs, dtype, symbols )

    elif isinstance( dtype, rdt.Struct ):
      return s.gen_port_struct_input( lhs, rhs, mangled_rhs, dtype, symbols )

    else:
      assert False, f"unrecognized data type {dtype}!"

  def gen_port_array_input( s, lhs, rhs, pnames, dtype, index, n_dim, symbols ):
    if not n_dim:
      return s.gen_port_input( lhs, rhs, pnames, dtype, symbols )
    else:
      set_comb, structs = [], []
      for idx in range( n_dim[0] ):
        _lhs = f"{lhs}[{idx}]"
        if index == 0:
          _rhs = f"{rhs}[{idx}]"
          _index = index-1
        else:
          _rhs = f"{rhs}"
          _index = index
        _set_comb, _structs = s.gen_port_array_input( _lhs, _rhs, pnames, dtype, _index, n_dim[1:], symbols )
        set_comb += _set_comb
        structs  += _structs
      return set_comb, structs

  def gen_comb_input( s, packed_ports, symbols ):
    set_comb, structs = [], []
    # Read all input ports ( except for 'clk' ) from component ports into
    # the verilated model. We do NOT want `clk` signal to be read into
    # the verilated model because only the sequential update block of
    # the imported component should manipulate it.

    for pnames, vname, rtype, port_idx in packed_ports:
      pnames_iter = cycle(pnames)
      p_n_dim, p_rtype = get_rtype( rtype )
      if s._get_direction( p_rtype ) == 'InPort' and pnames[0] != 'clk' and vname:
        dtype = p_rtype.get_dtype()
        lhs = "_ffi_m."+s._verilator_name(vname)
        rhs = "s.{}"
        idx = port_idx
        _set_comb, _structs = s.gen_port_array_input( lhs, rhs, pnames_iter, dtype, idx, p_n_dim, symbols )
        set_comb += _set_comb
        structs  += _structs

    return set_comb, structs

  #-------------------------------------------------------------------------
  # gen_comb_output
  #-------------------------------------------------------------------------

  def gen_port_vector_output( s, lhs, mangled_lhs, rhs, dtype, symbols ):
    dtype_nbits = dtype.get_length()
    blocks   = [ f's.{mangled_lhs} = Wire( {s._gen_bits_decl(dtype_nbits)} )',
                 '@s.update',
                 f'def osignal_{mangled_lhs}():',
                 f'  {lhs} = s.{mangled_lhs}' ]

    set_comb = s._gen_ref_read( 's.'+mangled_lhs, rhs, dtype_nbits )
    return set_comb, blocks

  def gen_port_struct_output( s, lhs, mangled_lhs, rhs, dtype, symbols ):
    dtype_nbits = dtype.get_length()
    # If the top-level signal is a struct, we add the datatype to symbol?
    dtype_name = dtype.get_class().__name__
    if dtype_name not in symbols:
      symbols[dtype_name] = dtype.get_class()

    blocks   = [ f's.{mangled_lhs} = Wire( {s._gen_bits_decl(dtype_nbits)} )',
                 '@s.update',
                 f'def ostruct_{mangled_lhs}():' ]

    # We create a long Bits object to accept CFFI value for struct
    # the temporary wire name

    # We create a new struct if we are copying values from verilator
    # world to pymtl land and send it out through the output of this
    # component
    upblk_content = [ f"{lhs} = {dtype_name}()" ]
    body, pos = s._gen_struct_write( 'o', lhs, 's.'+mangled_lhs, dtype, 0 )
    assert pos == dtype.get_length()

    upblk_content += body
    make_indent( upblk_content, 1 )

    blocks += upblk_content

    # We create a long Bits object tmp first
    # Then we load the full Bits to tmp
    set_comb = s._gen_ref_read( 's.'+mangled_lhs, rhs, dtype_nbits )
    return set_comb, blocks

  def gen_port_output( s, lhs, pnames, rhs, dtype, symbols ):
    lhs = lhs.format(next(pnames))

    mangled_lhs = s._pymtl_name_mangle( lhs )

    if isinstance( dtype, rdt.Vector ):
      return s.gen_port_vector_output( lhs, mangled_lhs, rhs, dtype, symbols )
    elif isinstance( dtype, rdt.Struct ):
      return s.gen_port_struct_output( lhs, mangled_lhs, rhs, dtype, symbols )
    else:
      assert False, f"unrecognized data type {dtype}!"

  def gen_port_array_output( s, lhs, pnames, rhs, dtype, index, n_dim, symbols ):
    if not n_dim:
      return s.gen_port_output( lhs, pnames, rhs, dtype, symbols )
    else:
      set_comb, structs = [], []
      for idx in range( n_dim[0] ):
        if index == 0:
          _lhs = f"{lhs}[{idx}]"
          _index = index-1
        else:
          _lhs = f"{lhs}"
          _index = index
        _rhs = f"{rhs}[{idx}]"
        _set_comb, _structs = s.gen_port_array_output( _lhs, pnames, _rhs, dtype, _index, n_dim[1:], symbols )
        set_comb += _set_comb
        structs  += _structs
      return set_comb, structs

  def gen_comb_output( s, packed_ports, symbols ):
    set_comb, structs = [], []
    for pnames, vname, rtype, port_idx in packed_ports:
      pnames_iter = cycle(pnames)
      p_n_dim, p_rtype = get_rtype( rtype )
      if s._get_direction( rtype ) == 'OutPort':
        dtype = p_rtype.get_dtype()
        lhs = "s.{}"
        rhs = "_ffi_m." + s._verilator_name(vname)
        idx = port_idx
        _set_comb, _structs = s.gen_port_array_output( lhs, pnames_iter, rhs, dtype, idx, p_n_dim, symbols )
        set_comb += _set_comb
        structs  += _structs
    return set_comb, structs

  #-------------------------------------------------------------------------
  # gen_line_trace_py
  #-------------------------------------------------------------------------

  def gen_line_trace_py( s, packed_ports ):
    """Return the line trace method body that shows all interface ports."""
    template = '{0}={{s.{0}}},'
    trace_string = ''
    for pnames, _, _, _ in packed_ports:
      for pname in pnames:
        trace_string += ' ' + template.format( pname )
    return f"      return f'{trace_string}'"

  #-------------------------------------------------------------------------
  # gen_internal_line_trace_py
  #-------------------------------------------------------------------------

  def gen_internal_line_trace_py( s, packed_ports ):
    """Return the line trace method body that shows all CFFI ports."""
    # Now that there could be multiple pnames that correspond to one vname,
    # I'm not sure how to generate internal line trace... maybe we should
    # deprecate internal_line_trace since it's not used by many anyways?
    # ret = [ '_ffi_m = s._ffi_m', 'lt = ""' ]
    # template = \
    #   "lt += '{vname} = {{}}, '.format(full_vector(s.{pname}, _ffi_m.{vname}))"
    # for pname, vname, port in packed_ports:
    #   if vname:
    #     pname = s._verilator_name(pname)
    #     vname = s._verilator_name(vname)
    #     ret.append( template.format(**locals()) )
    # ret.append( 'return lt' )
    # make_indent( ret, 2 )
    # return '\n'.join( ret )
    return "    return ''"

  #=========================================================================
  # Helper functions
  #=========================================================================

  def _verilator_name( s, name ):
    # TODO: PyMTL translation should generate dollar-sign-free Verilog source
    # code. Verify that this replacement rule here is not necessary.
    return name.replace('__', '___05F').replace('$', '__024')

  def _pymtl_name_mangle( s, name ):
    return name.replace('.', '_DOT_').replace('[', '_LB_').replace(']', '_RB_')

  def _get_direction( s, port ):
    if isinstance( port, rt.Port ):
      d = port.get_direction()
    elif isinstance( port, rt.Array ):
      d = port.get_sub_type().get_direction()
    else:
      assert False, f"{port} is not a port or array of ports!"
    if d == 'input':
      return 'InPort'
    elif d == 'output':
      return 'OutPort'
    else:
      assert False, f"unrecognized direction {d}!"

  def _get_c_n_dim( s, port ):
    if isinstance( port, rt.Array ):
      return port.get_dim_sizes()
    else:
      return []

  def _get_c_dim( s, port ):
    return "".join( f"[{i}]" for i in s._get_c_n_dim(port) )

  def _get_c_nbits( s, port ):
    if isinstance( port, rt.Array ):
      dtype = port.get_sub_type().get_dtype()
    else:
      dtype = port.get_dtype()
    return dtype.get_length()

  def _gen_ref_write( s, lhs, rhs, nbits ):
    if nbits <= 64:
      return [ f"{lhs}[0] = int({rhs})" ]
    else:
      ret = []
      ITEM_BITWIDTH = 32
      num_assigns = (nbits-1)//ITEM_BITWIDTH+1
      for idx in range(num_assigns):
        l = ITEM_BITWIDTH*idx
        r = l+ITEM_BITWIDTH if l+ITEM_BITWIDTH <= nbits else nbits
        ret.append( f"{lhs}[{idx}] = int({rhs}[{l}:{r}])" )
      return ret

  def _gen_ref_read( s, lhs, rhs, nbits ):
    if nbits <= 64:
      return [ f"{lhs} = Bits{nbits}({rhs}[0])" ]
    else:
      ret = []
      ITEM_BITWIDTH = 32
      num_assigns = (nbits-1)//ITEM_BITWIDTH+1
      for idx in range(num_assigns):
        l = ITEM_BITWIDTH*idx
        r = l+ITEM_BITWIDTH if l+ITEM_BITWIDTH <= nbits else nbits
        _nbits = r - l
        ret.append( f"{lhs}[{l}:{r}] = Bits{_nbits}({rhs}[{idx}])" )
      return ret

  def _gen_bits_decl( s, nbits ):
    if nbits <= 256:
      return f'Bits{nbits}'
    else:
      return f'mk_bits({nbits})'



from pymtl3.passes.backends.verilog import VerilatorImportConfigs

from .import_.VerilatorImportPass import VerilatorImportPass
from .translation.TranslationPass import TranslationPass
from .TranslationImportPass import TranslationImportPass



#=========================================================================
# PlaceholderPass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jan 27, 2020

from pymtl3 import Placeholder
from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.errors import PlaceholderConfigError


class PlaceholderPass( BasePass ):

  def __call__( s, m ):
    if not isinstance(m, Placeholder) and hasattr(m, 'config_placeholder'):
      raise PlaceholderConfigError(m,
          "the given object is not a Placeholder but has `config_placeholder` attribute!")

    if isinstance( m, Placeholder ):
      s.visit_placeholder( m )
    for child in m.get_child_components(repr):
      s.__call__( child )

  def visit_placeholder( s, m ):
    if not hasattr( m, '_placeholder_meta' ):
      m._placeholder_meta = PassMetadata()



from ..BasePass import BasePass
from ..sim.AddSimUtilFuncsPass import AddSimUtilFuncsPass
from ..sim.GenDAGPass import GenDAGPass
from ..sim.SimpleSchedulePass import SimpleSchedulePass
from ..sim.WrapGreenletPass import WrapGreenletPass
from ..tracing.CLLineTracePass import CLLineTracePass
from ..tracing.LineTraceParamPass import LineTraceParamPass
from .HeuristicTopoPass import HeuristicTopoPass
from .Mamba2020Pass import Mamba2020Pass
from .TraceBreakingSchedTickPass import TraceBreakingSchedTickPass
from .UnrollTickPass import UnrollTickPass


class UnrollSim( BasePass ):
  def __call__( s, top ):
    top.elaborate()
    GenDAGPass()( top )
    WrapGreenletPass()( top )
    SimpleSchedulePass()( top )
    UnrollTickPass()( top )
    top.lock_in_simulation()

class HeuTopoUnrollSim( BasePass ):
  def __call__( s, top ):
    top.elaborate()
    GenDAGPass()( top )
    WrapGreenletPass()( top )
    HeuristicTopoPass()( top )
    UnrollTickPass()( top )
    top.lock_in_simulation()

class TraceBreakingSim( BasePass ):
  def __call__( s, top ):
    top.elaborate()
    GenDAGPass()( top )
    WrapGreenletPass()( top )
    TraceBreakingSchedTickPass()( top )
    top.lock_in_simulation()

class Mamba2020( BasePass ):
  def __init__( s, line_trace=False ):
    s.line_trace = line_trace

  def __call__( s, top ):
    top.elaborate()
    GenDAGPass()( top )
    WrapGreenletPass()( top )
    if s.line_trace:
      CLLineTracePass()( top )
      LineTraceParamPass()( top )
    Mamba2020Pass()( top )
    AddSimUtilFuncsPass()( top )
    top.lock_in_simulation()
